from . import permutation_alignment
from .minimal_distortion_principle import minimal_distortion_principle
from .projection_back import projection_back

__all__ = ["permutation_alignment", "minimal_distortion_principle", "projection_back"]

PROJECTION_BACK_KEYWORDS = ["projection_back", "projection-back", "PB"]
MINIMAL_DISTORTION_PRINCIPLE_KEYWORDS = [
    "minimal_distortion_principle",
    "minimal-distortion-principle",
    "MDP",
]
